"""
Generates Kubernetes secrets based on a provided app name.

If the `templates` directory contains a secret called `oas-{app}-variables`, it
will check if that secret already exists in the cluster, and if not: generate
it. It does the same for an `oas-{app}-basic-auth` secret that will contain a
password as well as a htpasswd encoded version of it.

See https://open.greenhost.net/openappstack/openappstack/-/issues/891 for the
context why we use this script and not a helm chart to generate secrets.

usage: python generate_secrets.py template_filename.j2
"""

import crypt
import os
import secrets
import string
import sys
import yaml

import jinja2
import jinja2_base64_filters  # pylint: disable=unused-import
from kubernetes import client, config
from kubernetes.client.exceptions import ApiException
from kubernetes.utils import create_from_yaml


def main():
    """Run everything"""
    # Add jinja filters we want to use
    env = jinja2.Environment(extensions=["jinja2_base64_filters.Base64Filters"])
    env.filters["generate_password"] = generate_password

    if len(sys.argv) < 2:
        print("Please provide an app name as an argument")
        sys.exit(1)
    app_name = sys.argv[1]

    create_variables_secret(app_name, env)
    create_basic_auth_secret(app_name, env)


def get_templates_dir():
    """Returns directory that contains the Jinja templates used to create app
    secrets"""
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), 'templates')


def create_variables_secret(app_name, env):
    """Checks if a variables secret for app_name already exists, generates it if necessary"""
    variables_filename = \
        os.path.join(get_templates_dir(), f"oas-{app_name}-variables.yaml.jinja")
    if os.path.exists(variables_filename):
        # Check if k8s secret already exists, if not, generate it
        with open(variables_filename) as template_file:
            lines = template_file.read()
            secret_name, secret_namespace = get_secret_metadata(lines)
            if get_kubernetes_secret(secret_name, secret_namespace) is None:
                print(f"Adding secret {secret_name} in namespace"
                      f" {secret_namespace} to cluster.")
                template = env.from_string(lines)
                store_kubernetes_secret(template.render(), secret_namespace)
            else:
                print(f"Secret {secret_name} in namespace {secret_namespace}"
                      " already exists. Not generating new secrets.")


def create_basic_auth_secret(app_name, env):
    """Checks if a basic auth secret for app_name already exists, generates it if necessary"""
    basic_auth_filename = \
        os.path.join(get_templates_dir(), f"oas-{app_name}-basic-auth.yaml.jinja")
    if os.path.exists(basic_auth_filename):
        with open(basic_auth_filename) as template_file:
            lines = template_file.read()
            secret_name, secret_namespace = get_secret_metadata(lines)

            if get_kubernetes_secret(secret_name, secret_namespace) is None:
                basic_auth_username = 'admin'
                basic_auth_password = generate_password(32)
                basic_auth_htpasswd = gen_htpasswd(
                    basic_auth_username,
                    basic_auth_password)
                print(f"Adding secret {secret_name} in namespace"
                      f" {secret_namespace} to cluster.")
                template = env.from_string(
                        lines,
                        globals={
                            'pass': basic_auth_password,
                            'htpasswd': basic_auth_htpasswd
                })
                store_kubernetes_secret(template.render(), secret_namespace)
            else:
                print(f"Secret {secret_name} in namespace {secret_namespace}"
                      " already exists. Not generating new secrets.")

def get_secret_metadata(yaml_string):
    """Returns secret name and namespace from metadata field in a yaml string"""
    secret_info = yaml.safe_load(yaml_string)
    secret_name = secret_info['metadata']['name']
    # default namespace is flux-system, but other namespace can be
    # provided in secret metadata
    if 'namespace' in secret_info['metadata']:
        secret_namespace = secret_info['metadata']['namespace']
    else:
        secret_namespace = 'flux-system'
    return secret_name, secret_namespace


def get_kubernetes_secret(secret_name, namespace):
    """Returns the contents of a kubernetes secret or None if the secret does not exist."""
    try:
        secret = API.read_namespaced_secret(secret_name, namespace).data
    except ApiException as ex:
        # 404 is expected when the optional secret does not exist.
        if ex.status != 404:
            raise ex
        return None
    return secret

def store_kubernetes_secret(secret_string, namespace):
    """Converts secret_string into a yaml object and adds it to the cluster"""
    secret_yaml = yaml.safe_load(secret_string)
    api_client = client.api_client.ApiClient()
    api_response = create_from_yaml(
            api_client,
            yaml_objects=[secret_yaml],
            namespace=namespace)
    print(f"Secret created with api response: {api_response}")

def generate_password(length):
    """Generates a password of "length" characters"""
    length = int(length)
    password = ''.join((secrets.choice(string.ascii_letters) for i in range(length)))
    return password


def gen_htpasswd(user, password):
    """generate htpasswd entry for user with password"""
    return "{}:{}".format(user, crypt.crypt(
                password, crypt.mksalt(crypt.METHOD_SHA512)
            ),
        )

if __name__ == "__main__":
    config.load_kube_config()
    API = client.CoreV1Api()
    main()
